/* The load script for the tertiary bootloader itself */
   
INPUT( main.o SFBStartup.o SFBbody.o SFBtalk.o)
GROUP(-lhw -lreflex -lgcc -lc)

MEMORY
{
  vect         : o = 0, l = 4k                   /* Flash sector 0: trap vectors + misc setup */
  rom (rx)     : o = 0x2200, l = 0x10000-0x2200  /* Flash sector 1-8: Tertiary bootloader.. */
  fastram (rw) : o = 0x40000000, l = 32k         /* Interrupt vector, and user stack/RAM */
  bodyram (rw) : o = 0x7FD00000, l = 8k          /* 'body RAM' -- e.g., kinesthetic data -- in the USB RAM..*/
  talkram (rw) : o = 0x7FE00000, l = 16k         /* 'talk RAM' -- e.g., packet buffers -- in the Ethernet RAM */
}

SECTIONS
{

  /* code and constants */
  .text :
  {
    __text_start = ORIGIN(rom);

    /* Make sure our entry point section is included, next */
    KEEP(*(.text._sketchSetup))
    SFBStartup.o(.text .text.*)

    /* Now general code and constant segments */
    *(.text .text.*)
    *(.strings)

    /* Now all the C++ (etc) initialization codes,
       which will be run by __libc_csu_init (elf-init.c) */
    . = ALIGN(4);
    _init = .;
    KEEP(*(.init))

    . = ALIGN(4);
    __preinit_array_start = .;
    KEEP (*(.preinit_array))
    __preinit_array_end = .;

    . = ALIGN(4);
    __init_array_start = .;
    KEEP (*(SORT(.init_array.*)))
    KEEP (*(.init_array))
    __init_array_end = .;

    . = ALIGN(0x4);
    KEEP (*crtbegin.o(.ctors))
    KEEP (*(EXCLUDE_FILE (*crtend.o) .ctors))
    KEEP (*(SORT(.ctors.*)))
    KEEP (*crtend.o(.ctors))

    . = ALIGN(4);
    _fini = .;
    KEEP(*(.fini))

    . = ALIGN(4);
    __fini_array_start = .;
    KEEP (*(.fini_array))
    KEEP (*(SORT(.fini_array.*)))
    __fini_array_end = .;

    KEEP (*crtbegin.o(.dtors))
    KEEP (*(EXCLUDE_FILE (*crtend.o) .dtors))
    KEEP (*(SORT(.dtors.*)))
    KEEP (*crtend.o(.dtors))

   } >rom

  .ctors :
  {
    /* C++ ctor and dtor lists */
    __ctor_list__ = .;
    LONG((__ctor_end__ - __ctor_list__) / 4 - 2)
    *(SORT_BY_NAME(.ctors*))
    LONG(0)
    __ctor_end__ = .;
    __dtor_list__ = .;
    LONG((__dtor_end__ - __dtor_list__) / 4 - 2)
    *(SORT_BY_NAME(.dtors*))
    LONG(0)
    __dtor_end__ = .;

  } > rom

 .ARM.extab :
  {
     __exidx_start = .;
      *(.ARM.extab*)
      *(.gnu.linkonce.armextab.*)
  } > rom
 .ARM.exidx : 
  {
       *(.ARM.exidx*)
       *(.gnu.linkonce.armexidx.*) 
     __exidx_end = .;
  } > rom

  /* read-only data */
  .rodata :
  {
    __rodata_start = .;
    *(.rodata .rodata.*)  /* read-only data */
     __rodata_end = .;
     . = ALIGN(4);
     PROVIDE(__text_end = .); 
  } > rom

  /* BODYRAM initialization */
  . = ALIGN(4);
  __rom_bodyram_data_start = .;
  .bodyram : 
  {
    __ram_bodyram_data_start = .;

   /* bodyram gets regular .data declarations from SFBbody.o.. */
   SFBbody.o(.data .data*)

   /* ..plus stuff tagged for .bodyram, from anywhere */
    *(.bodyram .bodyram.*)

    __ram_bodyram_data_end = ALIGN(4);
  } > bodyram AT>rom

  .bodyrambss (NOLOAD) : 
  {
    __ram_bodyram_bss_start = ALIGN(4);
    SFBbody.o(.bss .bss.*)

   /* ..plus stuff tagged for .bss.bodyram, from anywhere */
    *(.bss.bodyram .bss.bodyram.*)

    __ram_bodyram_bss_end = ALIGN(4);
  } > bodyram 

  /* TALKRAM initialization */
  . = __rom_bodyram_data_start + SIZEOF(.bodyram);
  __rom_talkram_data_start = ALIGN(4);

  .talkram : 
  {
    __ram_talkram_data_start = ALIGN(4);

   /* talkram gets regular .data declarations from SFBtalk.o.. */
    SFBtalk.o(.data .data.*)

   /* ..plus stuff tagged for talkram, from anywhere */
    *(.talkram .talkram.*)

    __ram_talkram_data_end = ALIGN(4);
  } > talkram AT>rom

  .talkrambss (NOLOAD) : 
  {
    __ram_talkram_bss_start = ALIGN(4);
    SFBtalk.o(.bss .bss.*)

   /* ..plus stuff tagged for .bss.talkram, from anywhere */
    *(.bss.talkram .bss.talkram.*)

    __ram_talkram_bss_end = ALIGN(4);
  } > talkram 


  /* FASTRAM initialization */
  . = __rom_talkram_data_start + SIZEOF(.talkram);
  __rom_fastram_data_start = ALIGN(4);
  .fastram : 
  {
    __ram_fastram_data_start = ALIGN(4);

   /* First thing in fastram must be the interrupt table, which has to
      land at 0x40000000! */
   ASSERT (!(__ram_fastram_data_start - ORIGIN(fastram)), "Interrupt vector not at start of fastram");

   KEEP(*(.fastram.sfb_interrupt_vector.ram))

   /* fastram gets all regular .data declarations that haven't gone elsewhere.. */
    *(.data .data.*)

   /* ..plus stuff specifically tagged for fastram, from anywhere */
    *(.fastram .fastram.*)

    __ram_fastram_data_end = ALIGN(4);

  } > fastram AT>rom

  .footer :
  {
    KEEP(*(.footer .footer.*))
     . = ALIGN(4);
     PROVIDE(__rom_end = .); 
  } > rom

  __rom_length = __rom_end - __text_start;

  __ram_fastram_stack_top = ORIGIN(fastram)+LENGTH(fastram)-32; /* Avoid IAP ram usage (while maintaining stack alignment!) */

  .fastrambss (NOLOAD) : 
  {
    __ram_fastram_bss_start = ALIGN(4);
    *(.bss .bss.*)
    KEEP(*(.bssstackcanary .bssstackcanary.*))  /* At the very end of memory, a cookie to help detect stack overflow */
    __ram_fastram_bss_end = ALIGN(4);
  } > fastram AT>rom

}

